@@ -1,7 +1,7 @@
 # vim: ts=8:sts=8:sw=8:noexpandtab
 #
 # This file is part of ReText
-# Copyright: 2012-2015 Dmitry Shachnev
+# Copyright: 2012-2017 Dmitry Shachnev
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,6 +28,11 @@ from PyQt5.QtGui import QColor, QImage, QKeyEvent, QMouseEvent, QPainter, \
 QPalette, QTextCursor, QTextFormat, QWheelEvent
 from PyQt5.QtWidgets import QFileDialog, QLabel, QTextEdit, QWidget
 
+try:
+	from ReText.fakevimeditor import ReTextFakeVimHandler
+except ImportError:
+	ReTextFakeVimHandler = None
+
 colors = {
 	'marginLine': QColor(0xdc, 0xd2, 0xdc),
 	'currentLineHighlight': QColor(0xff, 0xff, 0xc8),
@@ -101,6 +106,8 @@ class ReTextEdit(QTextEdit):
 		self.document().blockCountChanged.connect(self.updateLineNumberAreaWidth)
 		self.cursorPositionChanged.connect(self.highlightCurrentLine)
 		self.document().contentsChange.connect(self.contentsChange)
+		if globalSettings.useFakeVim:
+			self.installFakeVimHandler()
 
 	def updateFont(self):
 		self.setFont(globalSettings.editorFont)
@@ -143,10 +150,7 @@ class ReTextEdit(QTextEdit):
 			return QTextEdit.contextMenuEvent(self, event)
 		oldcursor = self.textCursor()
 		cursor = self.cursorForPosition(event.pos())
-		pos = cursor.positionInBlock()
-		if pos == len(text):
-			pos -= 1
-		curchar = text[pos]
+		curchar = self.document().characterAt(cursor.position())
 		isalpha = curchar.isalpha()
 		cursor.select(QTextCursor.WordUnderCursor)
 		if not isalpha or (oldcursor.hasSelection() and
@@ -179,16 +183,19 @@ class ReTextEdit(QTextEdit):
 			documentIndentMore(self.document(), cursor)
 		elif key == Qt.Key_Backtab:
 			documentIndentLess(self.document(), cursor)
-		elif key == Qt.Key_Return and not cursor.hasSelection():
-			if event.modifiers() & Qt.ShiftModifier:
-				# Insert Markdown-style line break
-				markupClass = self.tab.getActiveMarkupClass()
-				if markupClass == MarkdownMarkup:
-					cursor.insertText('  ')
+		elif key == Qt.Key_Return:
+			markupClass = self.tab.getActiveMarkupClass()
 			if event.modifiers() & Qt.ControlModifier:
 				cursor.insertText('\n')
 				self.ensureCursorVisible()
+			elif self.tableModeEnabled and tablemode.handleReturn(cursor, markupClass,
+					newRow=(event.modifiers() & Qt.ShiftModifier)):
+				self.setTextCursor(cursor)
+				self.ensureCursorVisible()
 			else:
+				if event.modifiers() & Qt.ShiftModifier and markupClass == MarkdownMarkup:
+					# Insert Markdown-style line break
+					cursor.insertText('  ')
 				self.handleReturn(cursor)
 		else:
 			if event.text() and self.tableModeEnabled:
@@ -317,6 +324,13 @@ class ReTextEdit(QTextEdit):
 		else:
 			QTextEdit.insertFromMimeData(self, mimeData)
 
+	def installFakeVimHandler(self):
+		if ReTextFakeVimHandler:
+			fakeVimEditor = ReTextFakeVimHandler(self, self.parent)
+			fakeVimEditor.setSaveAction(self.parent.actionSave)
+			fakeVimEditor.setQuitAction(self.parent.actionQuit)
+			self.parent.actionFakeVimMode.triggered.connect(fakeVimEditor.remove)
+
 class LineNumberArea(QWidget):
 	def __init__(self, editor):
 		QWidget.__init__(self, editor)
